/*
 * ProFTPD: mod_rcs -- a module for using RCS for transparent version control
 *
 * Copyright (c) 2002-2003 TJ Saunders
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * This is mod_rcs, contrib software for proftpd 1.2.x and above.
 * For more information contact TJ Saunders <tj@castaglia.org>.
 *
 * $Id: mod_rcs.c,v 1.8 2004/05/13 15:00:38 tj Exp tj $
 */

#include "conf.h"
#include "privs.h"

#define MOD_RCS_VERSION "mod_rcs/0.8.1"

/* Make sure the version of proftpd is as necessary. */
#if PROFTPD_VERSION_NUMBER < 0x0001021001
# error "ProFTPD 1.2.10rc1 or later required"
#endif

static int rcs_engine = FALSE;
static uid_t rcs_userid = (uid_t) -1;
static gid_t rcs_groupid = (gid_t) -1;
static mode_t rcs_mode = (mode_t) -1;

/* NOTE: this module requires that the user in question has a valid
 * and executable shell
 */

/* NOTE: will this handle DefaultRoot?  Uploading into a directory that is
 *  not the current directory?
 */

/* necessary prototypes */
static int rcs_ssystem(const char *);

/* Support routines
 */

static int rcs_check_in(char *file) {
  int res = -1;
  char buf[256] = {'\0'}, mesg[256] = {'\0'};

  /* format the message to use as the log message */
  snprintf(mesg, sizeof(mesg), "*** " MOD_RCS_VERSION
    ": file uploaded via FTP by %s ***", session.user);
  mesg[sizeof(mesg) - 1] = '\0';

  /* format the command to be executed */
  snprintf(buf, sizeof(buf),
    "ci -u -m\"%s\" -t-\"Initial check-in by %s\" -w\"%s\" -q \"%s\"", mesg,
    session.user, session.user, file);
  buf[sizeof(buf) - 1] = '\0';

  res = rcs_ssystem(buf);

  PRIVS_ROOT
  if (pr_fsio_chown(file, rcs_userid, rcs_groupid) < 0)
    pr_log_debug(DEBUG0, MOD_RCS_VERSION ": error: unable to chown() %s: %s",
      file, strerror(errno));
  if (rcs_mode != (mode_t) -1 && pr_fsio_chmod(file, rcs_mode) < 0)
    log_debug(DEBUG0, MOD_RCS_VERSION ": error: unable to chmod() %s: %s",
      file, strerror(errno));
  PRIVS_RELINQUISH

  return WEXITSTATUS(res);
}

static int rcs_check_out(char *file) {
  int res = -1;
  char buf[256] = {'\0'};
  uid_t userid = session.uid;
  gid_t groupid = session.gid;

  /* format the command to be executed */
  snprintf(buf, sizeof(buf), "co -f -l -q \"%s\"", file);
  buf[sizeof(buf) - 1] = '\0';

  res = rcs_ssystem(buf);

  PRIVS_ROOT
  pr_fsio_chown(file, userid, groupid);
  PRIVS_RELINQUISH

  return WEXITSTATUS(res);
}

static int rcs_has_dir(char *file) {
  char rcs_dir[MAXPATHLEN] = {'\0'}, *tmp = NULL;

  memset(rcs_dir, '\0', sizeof(rcs_dir));
  sstrncpy(rcs_dir, file, sizeof(rcs_dir));

  /* If the requisite RCS subdirectory is not present, create it. I can't
   * use "./RCS" -- need to determine the directory from the file/path
   * given
   */

  if ((tmp = strrchr(rcs_dir, '/')) == NULL) {
    memset(rcs_dir, '\0', sizeof(rcs_dir));
    sstrncpy(rcs_dir, "./RCS", sizeof(rcs_dir));

  } else
    *tmp = '\0';

  if (mkdir(rcs_dir, 0755) < 0) {
    /* ignore EEXIST */
    if (errno != EEXIST) {
      pr_log_pri(PR_LOG_NOTICE, MOD_RCS_VERSION
        ": error: unable to create '%s': %s", rcs_dir, strerror(errno));
      return -1;
    }
  }

  return 0;
}

/* provides a "safe" version of the system(2) call by dropping all special
 * privileges, currently retained by the daemon, before exec() the given
 * command.
 */
static int rcs_ssystem(const char *cmd) {
  pid_t pid;
  int status;
  struct sigaction sa_ignore, sa_intr, sa_quit;
  sigset_t set_chldmask, set_save;

  if (!cmd)
    return 1;

  /* prepare signal dispositions */
  sa_ignore.sa_handler = SIG_IGN;
  sigemptyset(&sa_ignore.sa_mask);
  sa_ignore.sa_flags = 0;

  if (sigaction(SIGINT, &sa_ignore, &sa_intr) < 0)
    return -1;

  if (sigaction(SIGQUIT, &sa_ignore, &sa_quit) < 0)
    return -1;

  sigemptyset(&set_chldmask);
  sigaddset(&set_chldmask, SIGCHLD);

  if (sigprocmask(SIG_BLOCK, &set_chldmask, &set_save) < 0)
    return -1;

  if ((pid = fork()) < 0) {
    status = -1;

  } else if (pid == 0) {
    /* child */

    /* restore previous signal actions */
    sigaction(SIGINT, &sa_intr, NULL);
    sigaction(SIGQUIT, &sa_quit, NULL);
    sigprocmask(SIG_SETMASK, &set_save, NULL);

    /* drop all special privileges before exec()'ing the command.  This
     * allows for the user to specify arbitrary input via the given
     * filename without the admin worrying that some arbitrary command
     * is being executed that could take advantage of proftpd's retention
     * of root real user ID.
     */
    PRIVS_REVOKE

    execl("/bin/sh", "sh", "-c", cmd, NULL);

    /* execl() error */
    _exit(127);

  } else {
    /* parent */

    while (waitpid(pid, &status, 0) < 0)
      if (errno != EINTR) {
        status = -1;
        break;
      }
  }

  /* restore the previous signal actions */
  if (sigaction(SIGINT, &sa_intr, NULL) < 0)
    return -1;

  if (sigaction(SIGQUIT, &sa_quit, NULL) < 0)
    return -1;

  if (sigprocmask(SIG_SETMASK, &set_save, NULL) < 0)
    return -1;

  return status;
}

/* Configuration handlers
 */

/* usage: RCSEngine on|off */
MODRET set_rcsengine(cmd_rec *cmd) {
  config_rec *c = NULL;
  int bool = -1;
  
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_DIR);

  if ((bool = get_boolean(cmd, 1)) == -1)
    CONF_ERROR(cmd, "expected boolean parameter");

  c = add_config_param(cmd->argv[0], 1, (void *) bool);
  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

/* usage: RCSMode mode */
MODRET set_rcsmode(cmd_rec *cmd) {
  config_rec *c = NULL;
  char *tmp = NULL;
  mode_t tmp_mode;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_DIR);

  tmp_mode = (mode_t ) strtol(cmd->argv[1], &tmp, 8);

  if (tmp && *tmp)
    CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "'", cmd->argv[1],
      "' is not a valid mode", NULL));

  c = add_config_param(cmd->argv[0], 1, NULL);
  c->argv[0] = pcalloc(c->pool, sizeof(mode_t));
  *((mode_t *) c->argv[0]) = tmp_mode;
  c->flags |= CF_MERGEDOWN;
 
  return HANDLED(cmd);
}

/* usage: RCSOwner user group */
MODRET set_rcsowner(cmd_rec *cmd) {
  config_rec *c = NULL;

  CHECK_ARGS(cmd, 2);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL|CONF_DIR);

  /* defer name-to-ID lookup until later */
  c = add_config_param_str(cmd->argv[0], 2, cmd->argv[1], cmd->argv[2]);
  c->flags |= CF_MERGEDOWN;

  return HANDLED(cmd);
}

/* Command handlers
 */

MODRET rcs_post_stor(cmd_rec *cmd) {
  if (!rcs_engine)
    return DECLINED(cmd);

  if (rcs_check_in(cmd->arg) < 0)
    pr_log_debug(DEBUG0, MOD_RCS_VERSION ": unable to check file into RCS");

  rcs_engine = FALSE;

  return DECLINED(cmd);
}

MODRET rcs_post_stor_err(cmd_rec *cmd) {
  if (!rcs_engine)
    return DECLINED(cmd);

  /* If this is the case, check out the file again, to overwrite any
   * changes made to the file, then check it back in.
   */
  rcs_check_out(cmd->arg);
  rcs_check_in(cmd->arg);

  rcs_engine = FALSE;

  return DECLINED(cmd);
}

MODRET rcs_pre_stor(cmd_rec *cmd) {
  config_rec *c = NULL;
  mode_t *mode = NULL;

  if ((rcs_engine = get_param_int(CURRENT_CONF, "RCSEngine", FALSE)) != TRUE) {
    rcs_engine = FALSE;
    return DECLINED(cmd);
  }

  /* lookup any configured RCSUOwner and RCSMode values */
  if ((c = find_config(CURRENT_CONF, CONF_PARAM, "RCSOwner", FALSE)) == NULL) {
    rcs_userid = session.uid;
    rcs_groupid = session.gid;

  } else {
    struct passwd *pw = NULL;
    struct group *gr = NULL;
    char *username = (char *) c->argv[0], *groupname = (char *) c->argv[1];

    if (!strcmp(username, "~"))
      rcs_userid = session.uid;

    else if ((pw = auth_getpwnam(cmd->tmp_pool, username)) == NULL)
      pr_log_debug(DEBUG0, MOD_RCS_VERSION ": unknown RCSUserOwner '%s'",
        username);
    else
      rcs_userid = pw->pw_uid;

    if (!strcmp(groupname, "~"))
      rcs_groupid = gr->gr_gid;

    else if ((gr = auth_getgrnam(cmd->tmp_pool, groupname)) == NULL)
      pr_log_debug(DEBUG0, MOD_RCS_VERSION ": unknown RCSGroupOwner '%s'",
        groupname);
    else 
      rcs_groupid = gr->gr_gid;
  }


  if ((mode = (mode_t *) get_param_ptr(CURRENT_CONF, "RCSMode",
      FALSE)) == NULL)
    rcs_mode = (mode_t) -1;
  else
    rcs_mode = *mode;

  /* check for RCS subdirectory.  If engine is on, and subdirectory doesn't
   * exist, create it.
   */
  if (rcs_has_dir(cmd->arg) < 0) {
    rcs_engine = FALSE;
    return DECLINED(cmd);
  }

  /* Check out the given filename.  It is an error to check out a file
   * that is not yet under version control -- just ignore this error.
   */
  rcs_check_out(cmd->arg);
 
  return DECLINED(cmd);
}

/* Module API tables
 */

static conftable rcs_conftab[] = {
  { "RCSEngine",	set_rcsengine,		NULL },
  { "RCSMode",		set_rcsmode,		NULL },
  { "RCSOwner",		set_rcsowner,		NULL },
  { NULL }
};

static cmdtable rcs_cmdtab[] = {
  { PRE_CMD,	   C_STOR,	G_NONE,	rcs_pre_stor,	   FALSE, FALSE },
  { POST_CMD,	   C_STOR,	G_NONE,	rcs_post_stor,	   FALSE, FALSE },
  { POST_CMD_ERR,  C_STOR,	G_NONE,	rcs_post_stor_err, FALSE, FALSE },
  { 0, NULL }
};

module rcs_module = {
  NULL, NULL,

  /* Module API version 2.0 */
  0x20,

  /* the module name */
  "rcs",

  /* module configuration handler table */
  rcs_conftab,

  /* module command handler table */
  rcs_cmdtab,

  /* module authentication handler table */
  NULL,

  /* module initialization function */
  NULL,

  /* module "child mode" post-fork initialization function */
  NULL
};
